# -*- coding: utf-8; -*-
#
# (c) 2011 http://www.osinit.ru
#
# This file is part of Mandriva Management Console (MMC).
#
# MMC is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# MMC is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with MMC; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
import os.path
import re
import logging
import ldap.modlist
import commands
import xmlrpclib
import mmc.plugins.base
import time

from mmc.plugins.base import ldapUserGroupControl
from mmc.support import mmctools
from mmc.support.mmctools import ServiceManager, shlaunch
from mmc.support.config import PluginConfig


INI = "/etc/mmc/plugins/kerberos.ini"

VERSION = "2.5.3"
APIVERSION = "0:2:0"
REVISION = 20

def getVersion(): return VERSION
def getApiVersion(): return APIVERSION
def getRevision(): return REVISION

def activate():
    """
    This function define if "kerberos" module can be activated.
    @rtype: boolean
    @return: return True if it can be activated
    """
    config = KerberosConfig("kerberos")

    logger = logging.getLogger()

    if config.disabled:
        logger.info("kerberos plugin disabled by configuration.")
        return False

    try:
        ldapObj = ldapUserGroupControl()
    except ldap.INVALID_CREDENTIALS:
        logger.error("Can't bind to LDAP: invalid credentials")
        return False

    # Test if the Kerberos LDAP schema is available in the directory
    try:
         schema = ldapObj.getSchema("krbContainer")
         if len(schema) <= 0:
             logger.error("Kerberos schema is not included in LDAP directory");
             return False
    except:
        logger.exception("invalid schema")
        return False

    # Verify the realm
    realm = config.realm
    if not realm:
	logger.error("Can't get a kerberos realm. Please check that file %s contains a 'realm' option in the 'main' section" % INI)
        return False
    	
    
    # Verify if KDC init script exist
    init = config.kdcInit
    if not os.path.exists(init):
        logger.error(init + " does not exist")
        return False
    
    # Verify if Administration server init script exist
    init = config.kadminInit
    if not os.path.exists(init):
        logger.error(init + " does not exist")
        return False
    
    # Verify if kerberos krb5.conf file exist
    conf = config.krb5Conf
    if not os.path.exists(conf):
        logger.error(conf + " does not exist")
        return False
    
    # Verify if kerberos kdc.conf file exist
    conf = config.kdcConf
    if not os.path.exists(conf):
        logger.error(conf + " does not exist")
        return False
        
    # Check if realm entry is in LDAP directory
    realmContainer = KerberosLdapControl().getRealm()
    
    if not realmContainer:
            logger.error("Can't find krbRealmContainer entry in LDAP for realm %s. Please check your Kerberos LDAP configuration." % config.realm);
            return False
     
    return True

def kadminCommand(query):
    """
    Makes kadmin.local command 
    @param query: command with parameters
    @type query: str
    @rtype: str
    @return: full command line to be executed
    """
    return 'kadmin.local -q \"%s\"' % query

def getPrincipalsList(filter='*'):
    """
    Gets a list of principals
    @rtype: list
    @return: list of principals according to filter
    """
    result = []
    output = mmctools.shlaunch(kadminCommand('listprincs \\"%s\\"' % filter))
    for line in output:
	result.append(line[:line.index('@')])
    #1-st string contains information about authentication result
    result.pop(0)
    return result

def getPrincipal(name):
    """
    Gets full information about principal 
    @param name: principal's name
    @type name: str
    @rtype: dict
    @return: dictionnary of principal's parameters
    """
    result = {}
    output = mmctools.shlaunch(kadminCommand('getprinc \\"%s\\"' % name));
    for line in output:
	res = re.search('^(.+?):(.*)$', line)
	if res:
	    option = res.group(1).strip()
	    value = res.group(2).strip()
	    result[option] = value
    return result

def getPoliciesList(filter = ''):
    """
    Gets a list of policies
    @rtype: list
    @return: list of principals' password policies according to filter
    """
    result = []
    output = mmctools.shlaunch(kadminCommand('listpols \\"%s\\"' % filter))
    for line in output:
	result.append(line)
    #1-st string contains information about authentication result
    result.pop(0)
    return result

def getPasswordErrors(output):
    """
    Checks password errors messages in command's output 
    @param output: command's output 
    @type output: str
    @rtype: dict
    @return: password errors flags
    """
    result = {}
    result['minimum length error'] = checkParam(".* is too short .*", output)
    result['minimum classes error'] = checkParam(".* does not contain enough character classes .*", output)
    result['reuse error'] = checkParam(".* cannot reuse .*", output)
    return result


def changePrincipalPassword(name, isRandom, password = ''):
    """
    Changes the password of principal
    @param name: principal's name 
    @type name: str
    @param isRandom: if need to generate random password 
    @type isRandom: bool
    @param password: principal's password 
    @type password: str
    @rtype: dict
    @return: operation result and password errors flags
    """
    result = {}
    passwordParam = '-pw \\"%s\\"' % password;
    okResult = "Password for .* changed"
    if isRandom:
	passwordParam = '-randkey'
	okResult = "Key for .* randomized"
    output = commands.getoutput(kadminCommand('cpw %s \\"%s\\"' % (passwordParam, name))).split('\n')
    print output;
    changed = checkParam(okResult, output)
    result['changed'] = checkParam(okResult, output)
    if not changed:
	result['password errors'] = getPasswordErrors(output)
    return result

def addPrincipal(name, isRandomPassword, password = '', policy = '', expDate = 'never', passwordExpDate = 'never'):
    """
    Add a principal
    @param name: principal's name 
    @type name: str
    @param isRandomPassword: if need to generate random password 
    @type isRandomPassword: bool
    @param password: principal's password 
    @type password: str
    @param policy: principal's password policy
    @type policy: str
    @param expDate: principal's expiration date
    @type expDate: str
    @param passwordExpDate: principal's password expiration date
    @type passwordExpDate: str
    @rtype: dict
    @return: operation result and password errors flags
    """
    
    result = {}
    passwordParam = '-pw \\"%s\\"' % password
    if isRandomPassword:
	passwordParam = '-randkey'
    
    policyParam = ''
    if policy:
	policyParam = '-policy \\"%s\\"' % policy
    else:
	policyParam = '-clearpolicy'

    expParam = '-expire \\"%s\\"' % daylightlessDate(expDate)
    passwordExpParam = '-pwexpire \\"%s\\"' % daylightlessDate(passwordExpDate)

    query = 'addprinc %s %s %s %s \\"%s\\"' % (passwordParam, expParam, passwordExpParam, policyParam, name)
    output = commands.getoutput(kadminCommand(query)).split('\n');
    added = checkParam("Principal .* created", output)
    result['added'] = added
    if not added:
	result['password errors'] = getPasswordErrors(output)
    return result


def modifyPrincipal(name, isRandomPassword, password = '', policy = '', expDate = 'never', passwordExpDate = 'never'):
    """
    Modify principal's password and parameters
    @param name: principal's name 
    @type name: str
    @param isRandomPassword: if need to generate random password 
    @type isRandomPassword: bool
    @param password: principal's password 
    @type password: str
    @param policy: principal's password policy
    @type policy: str
    @param expDate: principal's expiration date
    @type expDate: str
    @param passwordExpDate: principal's password expiration date
    @type passwordExpDate: str
    @rtype: dict
    @return: operation result and password errors flags
    """
    result = {}
    passwordResult = True 
    if password or isRandomPassword:
	passwordResult = changePrincipalPassword(name, isRandomPassword, password)
	result['password state'] = passwordResult
	if not passwordResult['changed']:
	    result['modified'] = False;
	    return result
	 	
    changeAttrsResult = changePrincipalAttrs(name, policy, expDate, passwordExpDate)
    result['modified'] = changeAttrsResult
    return result

def daylightlessDate(date):
    daylightIndex = 8
    daylight = time.localtime()[daylightIndex]
    if date == "never" or not daylight:
	return date
    hour = 3600
    format = "%Y-%m-%d %H:%M:%S"
    dateTuple = time.strptime(date, format)
    newDateTuple = time.localtime(time.mktime(dateTuple) + hour)
    newDate = time.strftime(format,newDateTuple)
    #print "date is changed from %s to %s" % (date, newDate) 
    return newDate    

def changePrincipalAttrs(name, policy = '', expDate = 'never', passwordExpDate = 'never'):
    """
    Modify principal's parameters
    @param name: principal's name 
    @type name: str
    @param policy: principal's password policy
    @type policy: str
    @param expDate: principal's expiration date
    @type expDate: str
    @param passwordExpDate: principal's password expiration date
    @type passwordExpDate: str
    @rtype: bool
    @return: true if principal was modified else false
    """
    result = {}
	 	
    expParam = ''
    if expDate:
	expParam = '-expire \\"%s\\"' % daylightlessDate(expDate)
    
    passwordExpParam = ''
    if passwordExpDate:
	passwordExpParam = '-pwexpire \\"%s\\"' % daylightlessDate(passwordExpDate)
    
    policyParam = ""
    if policy:
	policyParam = '-policy \\"%s\\"' % policy
    else:
	princ = getPrincipal(name)
	if princ and princ['Policy'] != "[none]":
    	    policyParam = '-clearpolicy'
    
    query = 'modprinc %s %s %s \\"%s\\"' % (expParam, passwordExpParam, policyParam, name)
    print "query is %s" % kadminCommand(query)
    output = commands.getoutput(kadminCommand(query)).split('\n')
    print output
    localtime = time.localtime(time.time())
    print localtime
    return checkParam("Principal .* modified", output)
    
def changeKerberosPassword(name, password):
    """
    Change principal's password
    """
    return changePrincipalPassword(name, False, password)

def checkParam(param, output):
    """
    @param param: search string
    @type param: str
    @param otput: command's output
    @type output: str
    @rtype: bool
    @return: True if command's output contains search string
    """
    for line in output:
	#print line
        if re.search(param, line, flags = re.IGNORECASE):
    	    return True
    return False

def deletePrincipal(name):
    """
    Delete principal
    @rtype: bool
    @return: True if principal was deleted
    """
    output = commands.getoutput(kadminCommand('delprinc -force \\"%s\\"' % name)).split('\n')
    return checkParam("Principal .* deleted", output)

def isPrincipalExists(name):
    """
    @rtype: bool
    @return: True if principal exists
    """
    principals = getPrincipalsList(name + "*")
    searched = '^%s$' % name
    for p in principals:
	if re.search(searched, p):
	    return True
    return False

def getPolicy(name):
    """
    Gets full information about password policy 
    @param name: policy's name
    @type name: str
    @rtype: dict
    @return: dictionnary of policy's parameters
    """
    result = {}
    output = mmctools.shlaunch(kadminCommand('getpol \\"%s\\"' % name));
    for line in output:
	res = re.search('^(.+?):(.*)$', line)
	if res:
	    option = res.group(1).strip()
	    value = res.group(2).strip()
	    result[option] = value
    return result

def addPolicy(name, maxlife = 0, minlife = 0, minlength = 1, minclasses = 1):
    """
    Add password policy
    @param name: policy's name
    @type name: str
    @param maxlife: maximum lifetime of a password 
    @type maxlife: str
    @param minlife: minimum lifetime of a password 
    @type minlife: str
    @param minlength: minimum length of a password
    @type minlength: str
    @param minclasses: minimum number of character classes allowed in a password 
    @type minclasses: str
    @rtype: bool
    @return: True if policy was added
    """
    query = 'addpol -maxlife \\"%s seconds\\" -minlife \\"%s seconds\\" -minlength %s -minclasses %s \\"%s\\"' % (maxlife, minlife, minlength, minclasses, name)
    output = commands.getoutput(kadminCommand(query)).split('\n')
    return checkParam("Authenticating as principal .*", output) and (len(output) == 1)

def modifyPolicy(name, maxlife = 0, minlife = 0, minlength = 1, minclasses = 1):
    """
    Modify password policy
    @param name: policy's name
    @type name: str
    @param maxlife: maximum lifetime of a password 
    @type maxlife: str
    @param minlife: minimum lifetime of a password 
    @type minlife: str
    @param minlength: minimum length of a password
    @type minlength: str
    @param minclasses: minimum number of character classes allowed in a password 
    @type minclasses: str
    @rtype: bool
    @return: True if policy was modified
    """
    query = 'modpol -maxlife \\"%s seconds\\" -minlife \\"%s seconds\\" -minlength %s -minclasses %s \\"%s\\"' % (maxlife, minlife, minlength, minclasses, name)
    output = commands.getoutput(kadminCommand(query)).split('\n')
    return checkParam("Authenticating as principal .*", output) and (len(output) == 1)

def deletePolicy(name):
    """
    Delete password policy
    """
    result = {}
    result["broken list"] = breakPolicyLinks(name);
    output = commands.getoutput(kadminCommand('delpol -force \\"%s\\"' % name)).split('\n')
    result["deleted"]  = checkParam("Authenticating as principal .*", output) and (len(output) == 1)
    return result
    

def isPolicyExists(name):
    """
    @rtype: bool
    @return: True if policy exists
    """
    policies = getPoliciesList(name + "*")
    searched = '^%s$' % name
    for p in policies:
	if re.search(searched, p):
	    return True
    return False

def breakPolicyLinks(policyName):
    """
    Deletes links to password policy from principals
    """
    result = {}
    principalsNames = getPrincipalsList()
    for principalName in principalsNames:
	principal = getPrincipal(principalName)
	if principal['Policy'] == policyName:
	    modifyResult = modifyPrincipal(principalName,False, '', '', '', '')
	    result[principalName] = modifyResult['modified']
    return result

def getPrincipalsListWithPolicy(policy):
    """
    @rtype: list
    @return: all principals with this password policy
    """
    result = []
    principalsNames = getPrincipalsList()
    for principalName in principalsNames:
	principal = getPrincipal(principalName)
    	if principal["Policy"] == policy:
	    result.append(principalName)
    return result

def addKeytab(principal, path):
    """
    Adds an entry for principal to a keytab, randomizing principal's key in the process
    @param principal: principal's name
    @type principal: str
    @param path: path to keytab file
    @type path: str
    @rtype: bool
    @return: True if entry was added to keytab
    """
    output = commands.getoutput(kadminCommand('ktadd -k \\"%s\\" \\"%s\\"' % (path, principal))).split('\n')
    return checkParam(".* added to keytab .*", output) and os.path.exists(path)

def deleteKeytab(principal, path):
    """
    Removes entries for the specified principal from a keytab
    @param principal: principal's name
    @type principal: str
    @param path: path to keytab file
    @type path: str
    @rtype: bool
    @return: True if entry was removed from from keytab
    """
    result = {}
    result["file exists"] = os.path.exists(path)
    if result["file exists"]:
	output = commands.getoutput(kadminCommand('ktremove -k \\"%s\\" \\"%s\\" all' % (path, principal))).split('\n')
	result["deleted"] = checkParam(".* removed from keytab .*", output)
	result["no entry"] = checkParam("No entry .*", output)
    return result

def enablePrincipal(name):
    """
    Enables kerberos principal
    """
    output = commands.getoutput(kadminCommand('modprinc +allow_tix %s' % name)).split('\n');
    return checkParam("Principal .* modified", output)

def disablePrincipal(name):
    """
    Disables kerberos principal
    """
    output = commands.getoutput(kadminCommand('modprinc -allow_tix %s' % name)).split('\n');
    return checkParam("Principal .* modified", output)
    
def addKerberosAttributes(uid, password, policy, expDate, passExpDate):    
    """
    Adds kerberos principal for specified user
    """
    return KerberosLdapControl().addKerberosAttributes(uid, password, policy, expDate, passExpDate)

def isKerberosPrincipal(uid):
    """
    Checks if specified user has kerberos principal
    """
    return KerberosLdapControl().isKerberosPrincipal(uid)

def changeKerberosAttributes(uid, policy, expDate, passExpDate): 
    """
    Modifys attributes of kerberos principal for specified user
    """
    return KerberosLdapControl().changeKerberosAttributes(uid, policy, expDate, passExpDate)

def deleteKerberosAttributes(uid):
    """
    Deletes kerberos principal for specified user
    """
    return KerberosLdapControl().deleteKerberosAttributes(uid)    

def enableUser(uid):
    """
    Enables kerberos principal for specified user
    """
    return KerberosLdapControl().enableUser(uid)

def disableUser(uid):
    """
    Disables kerberos principal for specified user
    """
    return KerberosLdapControl().disableUser(uid)        

def kdcService(command):
    """
    Executes command for KDC service
    """
    return KDCService().command(command)

def kadminService(command):
    """
    Executes command for administration server (kadmind) 
    """
    return KadminService().command(command)

def isPrincipalForUser(principal):
    """
    Checks if specified principal was created for user
    """
    return KerberosLdapControl().isPrincipalForUser(principal)
    
class KerberosConfig(PluginConfig):
    """
    Gets configuration from kerberos.ini file
    """
    def readConf(self):
        PluginConfig.readConf(self)
        try:
    	    self.realm = self.get("main", "realm")
    	except: 
    	    pass
        self.kdcLockFile = self.get("kdc", "lockfile")
        self.kdcInit = self.get("kdc", "init")
        self.kadminLockFile = self.get("kadmin", "lockfile")
        self.kadminInit = self.get("kadmin", "init")
        self.krb5Conf = self.get("main", "krb5Conf")
        self.kdcConf = self.get("main", "kdcConf")
    
    def setDefault(self):
        PluginConfig.setDefault(self)
        self.realm = ""
        self.userDefault = {}
        
class KerberosServiceManager(ServiceManager):
    """
    Checks a kerberos service state
    """
    def isRunning(self):
        ret = False
        if os.path.exists(self.pidfile):
            ret = True
        return ret    
                                                                                                            
class KerberosLdapControl(mmc.plugins.base.ldapUserGroupControl):
    """
    Controls kerberos attributes for user via LDAP
    """
    def __init__(self, conffile = None, conffilebase = None):
        mmc.plugins.base.ldapUserGroupControl.__init__(self, conffilebase)
        self.configKerberos = KerberosConfig("kerberos", conffile)

    def addKerberosAttributes(self, uid, password, policy='', expDate='never', passExpDate='never'):
	"""
	Adds kerberos krbPrincipalAux objectclass and krbPrincipalName field to user's LDAP entry,
	creates kerberos principal for specified user and adds principal's name to user's krbPrincipalName field
    
	@param uid: user's uid corresponding to principal's name
	@type uid: str
	
	@param password: user's password
	@type password: str
	
	@param policy: kerberos password policy
	@type policy: str
	
	@param expDate: principal expiration date 
	@type expDate: str
	
	@param passExpDate: principal's password expiration date 
	@type passExpDate: str

	@rtype: dict
	@return: operation result and password errors flags
	"""
        # If the password has been encoded in the XML-RPC stream, decode it
        if isinstance(password, xmlrpclib.Binary):
            password = str(password)

        ret = addPrincipal(uid, False, password, policy, expDate, passExpDate)
    
	added = ret['added']
		
        if added:
	    # Command was successful, now set default attributes
            # Get current user entry
	    dn = 'uid=' + uid + ',' + self.baseUsersDN
	    s = self.l.search_s(dn, ldap.SCOPE_BASE)
	    c, old = s[0]
	    new = self._applyUserDefault(old, self.configKerberos.userDefault)

	    if not "krbPrincipalAux" in new["objectClass"]:
	      new["objectClass"].append("krbPrincipalAux")

	    krbPrincipalName = uid + '@' + self.configKerberos.realm
	    new["krbPrincipalName"] = krbPrincipalName

	    # Update LDAP
	    modlist = ldap.modlist.modifyModlist(old, new)
        
	    self.l.modify_s(dn, modlist)
      
        return ret

    def isKerberosPrincipal(self, uid):
        """
        @return: True if user has a Kerberos principal, else False
        @rtype: bool
        """
        ret = False
        if self.existUser(uid): 
#    	    print "User exists " + uid
    	    userinfo = self.getDetailedUser(uid)
    	    ret = ("krbPrincipalAux" in userinfo["objectClass"]) and ("krbPrincipalName" in userinfo) and (userinfo["krbPrincipalName"]!="")
    	return ret

    def changeKerberosAttributes(self, uid, policy='', expDate='never', passExpDate='never'):
        """
        Changes the attributes of Kerberos principal for specified user
    
        @param uid: login of the user
        @type uid: str
        
        @param policy: kerberos password policy
	@type policy: str
	
	@param expDate: principal expiration date 
	@type expDate: str
	
	@param passExpDate: principal's password expiration date 
	@type passExpDate: str

        @rtype: dict
        @return: operation result
        """
        dn = 'uid=' + uid + ',' + self.baseUsersDN
        s = self.l.search_s(dn, ldap.SCOPE_BASE)
        c, old = s[0]
        userinfo = self.getDetailedUser(uid)
        name = userinfo["krbPrincipalName"][0]

        result = {}        
    	result['modified'] = changePrincipalAttrs(name, policy, expDate, passExpDate)

        return result
        
    def deleteKerberosAttributes(self, uid):
        """
        Deletes Kerberos principal for specified user
        """
        userinfo = self.getDetailedUser(uid)
        name = userinfo["krbPrincipalName"][0]

        ret = deletePrincipal(name)
        return ret

    def enableUser(self, uid):
        """
        Enables Kerberos principal for specified user
        """
        userinfo = self.getDetailedUser(uid)
        name = userinfo["krbPrincipalName"][0]

        ret = enablePrincipal(name)
        return ret

    def disableUser(self, uid):
        """
        Disables Kerberos principal for specified user
        """
        userinfo = self.getDetailedUser(uid)
        name = userinfo["krbPrincipalName"][0]

        ret = disablePrincipal(name)
        return ret
        
    def getRealm(self):
        """
        Returns the LDAP entry corresponding to the current realm

        @return: the realm container entry
        @rtype: dict
        """
        realm = self.configKerberos.realm
        result = self.search("(&(objectClass=krbRealmContainer)(cn=%s))" % realm)
        if len(result): ret = result[0][0][1]
        else: ret = {}
        return ret


    def isPrincipalForUser(self, principal):
        """
        Check if principal was created for any user

        @return: checking result
        @rtype: bool
        """

	result = self.search("krbPrincipalName=%s@*" % principal,self.baseUsersDN)
	if len(result):
	    return True
	return False


class KDCService(KerberosServiceManager):
    """
    Controls KDC service
    """
    def __init__(self, conffile = None):
	self.config = KerberosConfig("kerberos", conffile)
        KerberosServiceManager.__init__(self, self.config.kdcLockFile, self.config.kdcInit)

class KadminService(KerberosServiceManager):
    """
    Controls kadmin service
    """
    def __init__(self, conffile = None):
	self.config = KerberosConfig("kerberos", conffile)
        KerberosServiceManager.__init__(self, self.config.kadminLockFile, self.config.kadminInit)
